#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <dirent.h>
#include <stdio.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/inotify.h>
#include <sys/param.h>

#include "io.h"
#include "trip.h"

char read_buffer[READ_BUFFER_SIZE];

char footer_string[] = "</div>\n"
                       "<hr>\n"
                       "<footer>\n"
                       "ZeroBBS " ZBBS_VERSION
                       "</footer>\n"
                       "</body>\n"
                       "</html>\n";

size_t readfile(char *fn) {
  int fd;
  size_t n;

  fd = open(fn, O_RDONLY);
  if (fd == -1)
    return 0;
  n = read(fd, read_buffer, READ_BUFFER_SIZE);
  close(fd);
  if (read_buffer[n-1] == '\n')
    n--;
  return n;
}

size_t readfileat(int dir, char *fn) {
  int fd;
  size_t n;

  fd = openat(dir, fn, O_RDONLY);
  if (fd == -1)
    return 0;
  n = read(fd, read_buffer, READ_BUFFER_SIZE);
  close(fd);
  if (read_buffer[n-1] == '\n')
    n--;
  return n;
}

void sleep_on(int dir, char *prop) {
  int in, fw;

  in = inotify_init();
  fw = inotify_add_watch(dir, prop, IN_DELETE_SELF|IN_MOVE_SELF);
  read(fw, read_buffer, READ_BUFFER_SIZE);
  close(fw);
  close(in);
}

size_t read_count(int board, char *p, int lock) {
  int old, new, carry, i;
  size_t n;
  char c;

  while ((new = openat(board, "wait", O_WRONLY|O_CREAT|O_EXCL, 0644)) == -1)
    sleep_on(board, "wait");
  old = openat(board, "count", O_RDONLY|O_CREAT, 0);
  n = read(old, p, READ_BUFFER_SIZE);
  if (n >= READ_BUFFER_SIZE-2)
    fail("System error: Counter overflow.\n");

  carry = 1;
  if (n) {
    n--;
    for (i=n-1; i>=0; i--) {
      if (p[i] == '9') {
        p[i] = '0';
        continue;
      }
      p[i] = p[i]+1;
      carry = 0;
      break;
    }
  }
  if (carry) {
    memmove(p+1, p, n);
    p[0] = '1';
    n++;
  }

  p[n] = '\n';
  write(new, p, n+1);
  if (!lock)
    renameat(board, "wait", board, "count");
  close(old);
  close(new);
  return n;
}

int bumpcmp(const void *e, const void *f) {
  const struct s_threadlist *te = e, *tf = f;

  if (te->zap != tf->zap)
    return (te->zap)?1:-1;
  if (te->pin != tf->pin)
    return (te->pin)?-1:1;
  if (te->time.tv_sec != tf->time.tv_sec)
    return tf->time.tv_sec - te->time.tv_sec;
  return tf->time.tv_nsec - te->time.tv_nsec;
}

size_t read_threads(DIR *dir, int board, struct s_threadlist *list, size_t nelem) {
  struct stat st;
  struct dirent *de;
  int thread;
  size_t n;

  n = 0;
  do {
    de = readdir(dir);
  } while (de > 0 && ( de->d_type != DT_DIR || de->d_name[0] == '.' ));
  while (de) {

    memcpy(list[n].id, de->d_name, 256);
    thread = openat(board, de->d_name, O_DIRECTORY|O_RDONLY);
    bzero(&st, sizeof(st));
    fstatat(thread, "bump", &st, AT_SYMLINK_NOFOLLOW);
    list[n].pin = getprop(thread, "pin"); 
    list[n].sage = getprop(thread, "sage"); 
    list[n].lock = getprop(thread, "lock"); 
    list[n].zap = getprop(thread, "zap");
    close(thread);
    list[n].time = st.st_mtim;
    if (++n == nelem)
      break;
    do {
      de = readdir(dir);
    } while (de > 0 && ( de->d_type != DT_DIR || de->d_name[0] == '.' ));

  }
  return n;
}

char hexbyte(char *p) {
  char c;

#define hex(v) ( (v) >= '0' && (v) <= '9' || (v) >= 'A' && (v) >= 'F' )
  if (!hex(p[0]) || !hex(p[1])) 
    return 0;
#undef hex
  return p[0]*16 + p[1] - '0'- 16*'0';
}

size_t html_escape(int out, char *s, char *p, char c, size_t n) {
  char *e;

  switch(c) {
    default:
    return n;

    case '&':
    if (p[0] == '#')
      return n;
    e = "&#38;";
    break;

    case '<':
    e = "&#60;";
    break;

    case '>':
    e = "&#62;";
    break;

    case '\'':
    e = "&#39;";
    break;

    case '"':
    e = "&#34;";
  }
  write(out, s+n, p-s-n-1);
  write(out, e, 5);
  return p-s;
}

void quote_string(int out, char *s) {
  char *p;
  size_t n;
  char c;

  p = s;
  n = 0;
  while (c = *p++)
    n = html_escape(out, s, p, c, n); 
  write(out, s+n, p-s-n-1);
}

size_t strrcspn(char *s, char *t, char *reject) {
  char *p;
  char c;

  for (c = *(p = t); p != s; c = *p--) {
    if (strchr(reject, c)) {
      p+=2;
      break;
    }
  }
  return t-p;
}

void write_markup(int out, char *s, char *number) {
  char *p, *q;
  int fd, quot;
  size_t n, i, j;
  char c;

  p = s;
  n = 0;
  quot = 0;
  while (c = *p++) {
    switch (c) {
      case '\n':
      write(out, s+n, p-s-n);
      if (quot)
        writestr(out, "</span>\n");
      quot = 0;
      writestr(out, "<br>\n");
      n = p-s;
      if (p[0] == '<') {
        writestr(out, "<span class='altquote'>");
        quot = 1;
      }
      break;

      case ':':
      if (p[0] != '/' || p[1] != '/')
        continue;
      i = strrcspn(s+n, p, " \n\t");
      j = 2 + strcspn(p+2, " \n\t");
      write(out, s+n, p-s-n-i);
      writestr(out, "<a href='");
      c = p[j];
      p[j] = 0;
      quote_string(out, p-i);
      writestr(out, "'>");
      quote_string(out, p-i);
      writestr(out, "</a>");
      p += j;
      *p = c;
      n = p-s;
      break;

      case '>':
      if (!quot) {
          q = 0;
          if (p[0] == '>') {
              i = strspn(p+1, "0123456789");
              if (!memchr(" \n\r\t", p[i+1], 5))
                goto nolink;
              j = strlen(number);
              if (i > j || i == j && strncmp(p+1, number, i) > 0)
                goto nolink;
            write(out, s+n, p-s-n-1);
            writestr(out, "<a href='#");
            write(out, p+1, i);
            writestr(out, "' class='link'>&#62;&#62;");
            write(out, p+1, i);
            writestr(out, "</a>");
            p += i+1;
            n = p-s;
              break;
          }
        nolink:
        if (p-1 == s || p[-2] == '\n') {
          write(out, s+n, p-s-n-1);
          writestr(out, "<span class='quote'>&#62;");
          n = p-s;
          quot = 1;
          break;
        }
      }

      default:
      n = html_escape(out, s, p, c, n);
      continue;
    }
  }
  write(out, s+n, p-s-n-1);
  if (quot)
    writestr(out, "</span>\n");
}

void write_name(int out, char *name) {
  char salt[16], pers[16];
  char *p, *q;
  int fd;
  size_t n;

  p = strchr(name, '#');
  if (!p) {
    quote_string(out, name);
    return;
  }

  fd = open(ZBBS_CONF "/capcode", O_RDONLY);
  if (fd != -1) {
    n = read(fd, read_buffer, READ_BUFFER_SIZE);
    close(fd);
    if (read_buffer[n-1] == '\n')
      n--;
    if (!strncmp(p+1, read_buffer, n)) {
      *p = 0;
      writestr(out, "<span class='post-capcode'>#</span>");
      quote_string(out, name);
      return;
    }
  }

  bzero(salt, 16);
  fd = open(ZBBS_CONF "/salt", O_RDONLY);
  read(fd, &salt[0], 16);
  close(fd);

  bzero(pers, 16);
  memcpy(pers, name, MIN(p-name, 16));

  *p = 0;
  n = 1 + strspn(p+1, "#");

  q = blake2b(p+n, salt, pers, n);
  quote_string(out, name);
  memset(p, '!', n);
  writestr(out, "<span class='post-trip'>");
  write(out, p, n);
  write(out, q, 12);
  writestr(out, "</span>");
}

void write_header(int out, char *title, char *boardname, int board) {
  struct dirent *de;
  DIR *dir;
  char *p;
  int fd, eof;
  ssize_t n, m;

  writestr(out,
           "<!DOCTYPE html>\n"
           "<html>\n"
           "<head>\n"
           "<title>\n");
  fd = open(ZBBS_CONF "/name", O_RDONLY);
  n = read(fd, read_buffer, READ_BUFFER_SIZE-1);
  close(fd);
  if (n < 0)
    n = 0;
  read_buffer[n] = 0;
  if (board)
    quote_string(out, read_buffer);
  writestr(out, "/");
  quote_string(out, boardname);
  writestr(out, "/\n");
  if (!board)
    quote_string(out, title);
  writestr(out, "\n</title>\n"
                "<meta charset='UTF-8'>\n"
                "<link rel='stylesheet' type='text/css' href='" WWW_CONF "/style.css'>\n"
                "<link rel='stylesheet' type='text/css' href='" WWW_CONF "/");
  quote_string(out, boardname);
  writestr(out, "'>\n"
                "<script src='" WWW_CONF "/script.js'></script>\n"
                "<link rel='icon' type='image/png' href='" WWW_CONF "/icon.png'>\n"
                "</head>\n"
                "<body>\n"
                "<div class='header'>\n"
                "<a href='index.html'>");
  quote_string(out, read_buffer);
  writestr(out, "</a>\n");

  writestr(out, "[\n");
  dir = opendir(".");
  do {
    de = readdir(dir);
  } while (de > 0 && ( de->d_type != DT_DIR || de->d_name[0] == '.' || access(de->d_name, W_OK) ));
  if (de > 0) for (;;) {
    writestr(out, "<a href='/");
    quote_string(out, de->d_name);
    writestr(out, "'>");
    quote_string(out, de->d_name);
    writestr(out, "</a>\n");
    do {
      de = readdir(dir);
    } while (de > 0 && ( de->d_type != DT_DIR || de->d_name[0] == '.' || access(de->d_name, W_OK) ));
    if (de > 0)
      writestr(out, "/\n");
    else
      break;
  }
  closedir(dir);
  writestr(out, "]\n</div>\n<hr>\n");
  fd = open(ZBBS_CONF "/banner.html", O_RDONLY);
  while ((n = read(fd, read_buffer, READ_BUFFER_SIZE)) > 0)
    write(out, read_buffer, n);
  close(fd);
}

void write_post(int out, char *number, char *name, char *subject,
                char *mail, char *body) {
  time_t date;
  size_t n;

  writestr(out, "<div class='post' id='");
  n = strlen(number);
  write(out, number, n);
  writestr(out, "'>\n"
                "<h3 class='post-info'>\n"
                "<span class='post-number'>");
  write(out, number, n);
  writestr(out, "</span>\n"
                "<span class='post-subject'>");
  quote_string(out, subject);

  if (mail) {
    writestr(out, "</span>\n"
                  "<a href='mailto:");
    quote_string(out, mail);
    writestr(out, "'><span class='post-name'>");
    write_name(out, name);
    writestr(out, "</span></a>\n"
                  "<span class='post-date'>");

  }

  else {
    writestr(out, "</span>\n"
                  "<span class='post-name'>");
    write_name(out, name);
    writestr(out, "</span>\n"
                  "<span class='post-date'>");
  }
  date = time(NULL);
  n = strftime(read_buffer,
               READ_BUFFER_SIZE,
               "%Y-%m-%d %H:%M:%S",
               localtime(&date));
  write(out, read_buffer, n);
  writestr(out, "</span>\n"
                "</h3>\n"
                "<div class='post-body'>\n"
                "<blockquote>\n");
  write_markup(out, body, number);
  writestr(out, "\n</blockquote>\n"
                "</div>\n"
                "</div>\n");
}

void write_board(int out, int board, char *name, struct s_threadlist *list, size_t nelem) {
  int thread, fd;
  size_t i, n;

  writestr(out, "<div class='thread-creation'>\n"
                "<details>\n"
                "<summary><h3>[ Create new thread ]</h3></summary>\n"
                "<hr>\n"
                "<form method='post' action='" WWW_BIN "/http-post-thread'>\n"
                "<input type='hidden' name='board' value='");
  write(out, name, strlen(name));
  writestr(out, "'>\n"
                "<table class='thread-form'>\n"
                "<tr>\n"
                "<th class='form-field'>Title</th>\n"
                "<td>"
                "<input type='text' name='title' autocomplete='off' class='form-entry'>\n"
                "<input type='submit' value='Post' class='form-button'>\n"
                "</td>\n"
                "</tr>\n"
                "<tr>\n"
                "<th class='form-field'>Name</th>\n"
                "<td>"
                "<input type='text' name='name' autocomplete='off' class='form-entry'>\n"
                "</td>\n"
                "</tr>\n"
                "<tr>\n"
                "<th class='form-field'>Mail</th>\n"
                "<td>"
                "<input type='text' name='mail' autocomplete='off' class='form-entry'>\n"
                "</td>\n"
                "</tr>\n"
                "<tr>\n"
                "<th class='form-field'>Subject</th>\n"
                "<td>"
                "<input type='text' name='subject' autocomplete='off' class='form-entry'>\n"
                "</td>\n"
                "</tr>\n"
                "<tr>\n"
                "<th class='form-field'>Comment</th>\n"
                "<td>"
                "<textarea name='comment' autocomplete='off' rows='10' cols='50' class='form-entry'></textarea>\n"
                "</td>\n"
                "</tr>\n"
                "</table>\n"
                "</form>\n"
                "<hr>\n"
                "</details>\n"
                "</div>\n"
                "<div>\n"
                "<fieldset>\n"
                "<legend class='board-title'>/");
  quote_string(out, name);
  writestr(out, "/</legend>\n"
                "<span class='board-nav'><a href='.'>[Refresh]</a></span>\n"
                "<ul class='board-threads'>\n");
  for (i=0; i!=nelem; i++) {
    writestr(out, "<li class='board-thread'>\n");
    if (list[i].zap)
      writestr(out, "<span class='zapped-thread'>");
    else {
      writestr(out, "<a href='");
      write(out, list[i].id, strlen(list[i].id));
      if (list[i].pin)
        writestr(out, "' class='pinned-thread'");
      else if (list[i].sage)
        writestr(out, "' class='saged-thread'");
      writestr(out, "'>\n");
    }
    write(out, list[i].id, strlen(list[i].id));
    writestr(out, ": ");
    thread = openat(board, list[i].id, O_DIRECTORY|O_RDONLY);
    n = readfileat(thread, "title");
    if (n)
      write(out, read_buffer, n);
    else
      writestr(out, "Untitled");
    if (list[i].zap)
      writestr(out, "</span>\n");
    else
      writestr(out, "</a>\n");
    n = readfileat(thread, "comment");
    if (n) {
      writestr(out, "<span class='thread-comment'>(");
      write(out, read_buffer, n);
      writestr(out, ")</span>\n");
    }
    writestr(out, "</li>\n");
  }
  writestr(out, "</ul>\n"
                "<span>");
  n = readfileat(board, "count");
  if (!n)
    writestr(out, "0 threads created</span>\n"
                  "</fieldset>\n");
  else if (n == 1 && read_buffer[0] == '1')
    writestr(out, "1 thread created</span>\n"
                  "</fieldset>\n");
  else {
    write(out, read_buffer, n);
    writestr(out, " threads created</span>\n"
                  "</fieldset>\n");
  }
}

void write_thread(int out, char *board, char *number, char *title,
                  char *name, char *subject, char *mail, char *body) {
  writestr(out, "<div class='thread-creation'>\n"
                "<form method='post' action='" WWW_BIN "/http-post-reply'>\n"
                "<input type='hidden' name='board' value='");
  write(out, board, strlen(board));
  writestr(out, "'>\n"
                "<input type='hidden' name='thread' value='");
  write(out, number, strlen(number));
  writestr(out, "'>\n"
                "<table class='thread-form'>\n"
                "<tr>\n"
                "<th class='form-field'>Name</th>\n"
                "<td>"
                "<input type='text' name='name' autocomplete='off' class='form-entry'>\n"
                "<input type='submit' value='Reply' class='form-button'>\n"
                "</td>\n"
                "</tr>\n"
                "<tr>\n"
                "<th class='form-field'>Mail</th>\n"
                "<td>"
                "<input type='text' name='mail' autocomplete='off' class='form-entry'>\n"
                "</td>\n"
                "</tr>\n"
                "<tr>\n"
                "<th class='form-field'>Subject</th>\n"
                "<td>"
                "<input type='text' name='subject' autocomplete='off' class='form-entry'>\n"
                "</td>\n"
                "</tr>\n"
                "<tr>\n"
                "<th class='form-field'>Comment</th>\n"
                "<td>"
                "<textarea name='comment' autocomplete='off' rows='10' cols='50' class='form-entry'></textarea>\n"
                "</td>\n"
                "</tr>\n"
                "</table>\n"
                "</form>\n"
                "<hr>\n"
                "</div>\n"
                "<div class='thread'>\n"
                "<span class='thread-number'>[");
  write(out, number, strlen(number));
  writestr(out, "]</span>\n"
                "<span class='thread-title'>");
  quote_string(out, title);
  writestr(out, "</span>\n"
                "<span class='thread-nav'>\n"
                "<a href=''>[Refresh]</a>\n"
                "<a href='..'>[Return]</a>\n"
                "</span>\n"
                "<hr>\n");
  write_post(out, "1", name, subject, mail, body);
}

void write_footer(int out) {
  writestr(out, footer_string);
}

void seek_footer(int out) {
  lseek(out, 1-sizeof(footer_string), SEEK_END);
}

int check_footer(int in) {
 size_t n;

  if (lseek(in, 1-sizeof(footer_string), SEEK_END) < 0)
    return 1;
  n = read(in, read_buffer, sizeof(footer_string));
  if (memcmp(footer_string, read_buffer, sizeof(footer_string)))
    return 1;
  return 0;
}
