#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "util.h"

char read_buffer[READ_BUFFER_SIZE];

#define hex(v) ((v) >= '0' && (v) <= '9' || (v) >= 'A' && (v) <= 'F')
char hexbyte(char *p) {
  char i;

  if (p[0] >= '0' && p[0] <= '9')
    i = p[0] - '0';
  else if (p[0] >= 'A' && p[0] <= 'F')
    i = p[0] + 10 - 'A';
  else
    return 0;
  i *= 16;

  if (p[1] >= '0' && p[1] <= '9')
    i += p[1] - '0';
  else if (p[1] >= 'A' && p[1] <= 'F')
    i += p[1] + 10 - 'A';
  else
    return 0;

  return i;
}
#undef hex

size_t urldecode(char *s, size_t len) {
  char *p;
  size_t i, n, m;
  char c;

  n = 0;
  m = 0;
  for (i=1; i!=len && s[i] != '&' ; i++) {
    switch(s[i]) {
      case '+':
      memmove(s+n-m, s+n, i-n);
      s[i-m] = ' ';
      n = i+1;
      break;

      case '%':
      c = hexbyte(s+i+1);
      if (!c) /* never send null bytes */
        continue;
      memmove(s+n-m, s+n, i-n);
      s[i-m] = c;
      n = i+3;
      p += 2;
      m += 2;
      break;
    
      case ' ': /* strip whitespace */
      case '\n':
      case '\r':
      case '\t':
      memmove(s+n-m, s+n, i-n);
      n = i+1;
      m += 1;
    }
  }
  memmove(s+n-m, s+n, i-n);
  s[i-m] = 0;
  return i;
}

int findstr(char *s, char **list, size_t n) {
  size_t i;

  for(i=0; i!=n; i++) {
    if (!strcmp(s, list[i]+1))
      return 1;
  }
  return 0;
}

int main(int argc, char **argv, char **envp) {
  char *p;
  ssize_t n;
  int mask, unquoted;
  size_t m, i, block, skip;
  char c;

  if (argv[1][0] == '-') {
    mask = 1;
    skip = 2;
  }
  else {
    mask = 0;
    skip = 1;
  }

  unquoted = 0;
  for (block=skip; block!=argc && argv[block][0]; block++)
    if (argv[block][0] != ' ')
      unquoted = 1;

  if (unquoted || block == argc || block < 2 ||
      argv[1][0] == '-' &&
       ( argv[1][1] != 'c' || argv[1][2]))
    fail("Usage: urlformenv { key [...] | -c [key ...] } prog ...\n");

  n = read(0, read_buffer, READ_BUFFER_SIZE-1);
  if (n > 0) {
    if (n == READ_BUFFER_SIZE-1)
      fail("Message exceeds size limit.\n");

    m = 0;
    do {
      p = memchr(read_buffer+m, '=', n);
      if (!p)
        fail("Malformed url parameter.\n");
      *p = 0;
      i = urldecode(p, read_buffer+n-p);
      if (mask ^ findstr(read_buffer+m, argv+skip, block-1))
        setenv(read_buffer+m, p+1, 1);
      m = p+i+1-read_buffer;
    } while(m != n+1);
  }

  execvp(argv[block+1], argv+block+1);
  return 0;
}
