#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <stdio.h>
#include <sys/wait.h>

#include "io.h"

void fix_wait(int dir, char *board, char *file) {
  if (!getprop(dir, "wait"))
    return;
  if (faccessat(dir, "count", F_OK, AT_SYMLINK_NOFOLLOW))
    renameat(dir, "wait", dir, "count");
  else
    unlinkat(dir, "wait", 0);
  writestr(1, "removed writelock ");
  write(1, board, strlen(board));
  if (file) {
    writestr(1, "/");
    write(1, file, strlen(file));
  }
  writestr(1, "/wait\n");
}

int main(int argc, char **argv, char **envp) {
  char *call[4] = { ZBBS_BIN "/zbbs-unthread", NULL, NULL, NULL };
  struct s_threadlist list[THREAD_LIMIT*2];
  DIR *dir;
  int fd, fe, board, i;
  size_t n, j;

  if (argc == 1)
    fail("Usage: zbbs-fsck board [...]\n");

  for(i=1; i!=argc; i++) {
    board = open(argv[i], O_RDONLY);
    if (board == -1) {
      writestr(1, "Board ");
      write(1, argv[i], strlen(argv[i]));
      writestr(1, "does not exist.\n");
      continue;
    }
    fix_wait(board, argv[i], NULL);

    dir = fdopendir(dup(board));
    n = read_threads(dir, board, &list[0], THREAD_LIMIT*2);
    if (n > THREAD_LIMIT) {
      qsort(list, n, sizeof(struct s_threadlist), &bumpcmp);
      n -= THREAD_LIMIT;
      call[1] = argv[i];
      dup2(1, 2);
      for (;;) {
        for (j=0; j!=n; j++)
          if (!fork()) {
            call[2] = list[THREAD_LIMIT+j].id;
            execve(call[0], call, envp);
          }
          else
            wait(NULL);
        if (n != THREAD_LIMIT)
          break;
        n = read_threads(dir, fd, &list[THREAD_LIMIT], THREAD_LIMIT);
        qsort(list, n+THREAD_LIMIT, sizeof(struct s_threadlist), &bumpcmp);
      }
      n = THREAD_LIMIT;
    }
    closedir(dir);

    for (j=0; j!=n; j++) {
      fd = openat(board, list[j].id, O_DIRECTORY|O_RDONLY);
      if (fd != -1) {
        fix_wait(fd, argv[i], list[j].id);
        fe = openat(fd, "index.html", O_RDONLY);
        if (fe == -1 || check_footer(fe)) {
          writestr(1, "found damaged content at ");
          write(1, argv[i], strlen(argv[i]));
          writestr(1, "/");
          write(1, list[j].id, strlen(list[j].id));
          writestr(1, "\n");
        }
        if (fe != -1)
          close(fe);
      }
      close(fd);
    }
    close(board);
  }
}
