#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>

#include "io.h"

int main(int argc, char **argv, char **envp) {
  char id[READ_BUFFER_SIZE];
  char title_buffer[] = "Untitled";
  char *title, *name, *mail;
  int board, thread, out;
  size_t n, m;

  if (argc != 7)
    fail("Usage: zbbs-thread board title name mail subject comment\n");
  if (strchr(argv[1], '/'))
    fail("Invalid character in board name.\n");
  if (!argv[6][0])
    fail("Post empty.\n");
  title = (argv[2][0])?argv[2]:title_buffer;
  name = (argv[3][0])?argv[3]:"Anonymous";
  mail = (argv[4][0])?argv[4]:NULL;

  board = open(argv[1], O_RDONLY|O_DIRECTORY);
  if (board == -1 || access(argv[1], W_OK)) /* confirm writable */
    fail("Board does not exist.\n");
  else if (getprop(board, "lock"))
    fail("Board is locked.\n");
  
  do {
    n = read_count(board, id, 0);
    id[n] = 0;
  } while (mkdirat(board, id, 0750));
  thread = openat(board, id, O_DIRECTORY|O_RDONLY);
  close(openat(thread, "wait", O_CREAT|O_RDONLY, 0666));

  m = strlen(title);
  title[m] = '\n';
  out = openat(thread, "title", O_WRONLY|O_CREAT, 0666);
  write(out, title, m+1);
  close(out);
  title[m] = 0;

  out = openat(thread, "index.html", O_WRONLY|O_CREAT, 0666);
  write_header(out, title, argv[1], 0);
  write_thread(out, argv[1], id, title, name, argv[5], mail, argv[6]);
  write_footer(out);

  out = openat(thread, "count", O_WRONLY|O_CREAT, 0666);
  writestr(out, "1\n");
  close(out);

  close(openat(thread, "bump", O_CREAT|O_RDONLY, 0666));
  if (fork()) {
    unlinkat(thread, "wait", 0);
    close(out);
    close(thread);
    close(board);
    id[n] = '\n';
    write(1, id, n+1);
    writestr(2, "new thread ");
    write(2, argv[1], strlen(argv[1]));
    writestr(2, "/");
    write(2, id, n+1);
    exit(0);
  }
  argv[0] = ZBBS_BIN "/zbbs-board";
  argv[2] = 0; /* pass board */
  execve(argv[0], argv, envp);
  return 0;
}
